# pylint: disable=too-many-public-methods
import os
import abc
import json
import types
import typing
import random
import logging
import tempfile

import cv2
import torch
import tqdm
import numpy as np
import timm.optim
import timm.scheduler
import pkg_resources
import typing_extensions as tx

try:
    import model_archiver.model_packaging as marmp
    import model_archiver.model_packaging_utils as marmpu
except ImportError:
    marmp = None
    marmpu = None

from .. import metrics as mm
from .. import core as mc
from . import common as mdc

DEFAULT_SCHEDULER_PARAMS = dict(
    sched="cosine",
    min_lr=1e-5,
    decay_rate=1,
    warmup_lr=0,
    warmup_epochs=0,
    cooldown_epochs=0,
    epochs=100,
    lr_cycle_limit=0,
)

DEFAULT_OPTIMIZER_PARAMS = dict(learning_rate=1e-2, weight_decay=4e-5)

ResizeMethod = tx.Literal["fit", "pad"]

LOGGER = logging.getLogger(__name__)


class Detector:
    """Abstract base class for a detector."""

    @property
    @abc.abstractmethod
    def anchor_boxes(self) -> np.ndarray:
        """Return the list of anchor boxes in xyxy format."""

    model: torch.nn.Module
    backbone: torch.nn.Module
    annotation_config: mc.AnnotationConfiguration
    training_model: torch.nn.Module
    device: typing.Any
    resize_method: tx.Literal["pad", "fit", "pad_to_multiple"]
    resize_base: typing.Optional[int]

    def set_device(self, device):
        """Set the device for training and inference tasks."""
        self.device = torch.device(device)
        self.model.to(self.device)
        if hasattr(self, "training_model"):
            self.training_model.to(self.device)  # type: ignore

    @abc.abstractmethod
    def invert_targets(
        self,
        y: typing.Any,
        threshold: float = 0.5,
        **kwargs,
    ) -> typing.List[typing.List[mc.Annotation]]:
        """Compute a list of annotation groups from model output."""

    def resize_to_model_size(
        self, images: typing.List[np.ndarray]
    ) -> typing.Tuple[np.ndarray, np.ndarray]:
        """Resize a series of images to the current model's size."""
        height, width = self.input_shape[:2]
        padded, scales, _ = mdc.resize(
            images,
            resize_method=self.resize_method,
            height=height,
            width=width,
            base=getattr(self, "resize_base", None),
        )
        # Ensure that scaling maintains aspect ratio.
        np.testing.assert_allclose(*scales.T)
        return padded, scales[:, 0]

    @property
    @abc.abstractmethod
    def input_shape(self) -> typing.Tuple[int, int, int]:
        """Obtain the input shape for this model."""

    @abc.abstractmethod
    def set_input_shape(self, width: int, height: int):
        """Set the input shape for this model."""

    @abc.abstractmethod
    def serve_module_string(self, enable_flexible_size: bool) -> str:
        """Return the module string used as part of TorchServe."""

    @property
    @abc.abstractmethod
    def serve_module_index(self) -> dict:
        """Return the class index -> label mapping for TorchServe."""

    @abc.abstractmethod
    def compute_inputs(self, images: np.ndarray) -> np.ndarray:
        """Convert images into suitable model inputs. *You
        usually should not need this method*. For training,
        use `detector.train()`. For detection, use
        `detector.detect()`.

        Args:
            images: The images to convert

        Returns:
            The input to the model
        """

    @abc.abstractmethod
    def compute_targets(
        self,
        annotation_groups: typing.List[typing.List[mc.Annotation]],
        width: int,
        height: int,
    ) -> typing.Union[typing.List[np.ndarray], np.ndarray]:
        """Compute the expected outputs for a model. *You
        usually should not need this method*. For training,
        use `detector.train()`. For detection, use
        `detector.detect()`.

        Args:
            annotation_groups: A list of lists of annotation groups.

        Returns:
            The output(s) that will be used by detector.train()
        """

    def freeze_backbone(self):
        """Freeze the body of the model, leaving the final classification and
        regression layer as trainable."""
        for p in self.backbone.parameters():  # type: ignore
            p.requires_grad = False
        for m in self.backbone.modules():  # type: ignore
            m.eval()

    def unfreeze_backbone(self, batchnorm=True):
        """Unfreeze the body of the model, making all layers trainable.

        Args:
            batchnorm: Whether to unfreeze batchnorm layers.
        """
        for m in self.backbone.modules():  # type: ignore
            if isinstance(m, torch.nn.BatchNorm2d) and not batchnorm:
                m.eval()
                for p in m.parameters():
                    p.requires_grad = False
            else:
                m.train()
                for p in m.parameters():
                    p.requires_grad = True

    def loss_for_batch(
        self,
        batch,
        data_dir: str = None,
        transforms: np.ndarray = None,
        indices: typing.List[int] = None,
    ):
        """Compute the loss for a batch of scenes."""
        assert self.training_model.training, "Model not in training mode."
        images, scales = self.resize_to_model_size(batch.images)
        annotation_groups = [
            [ann.resize(scale) for ann in scene.annotations]
            for scene, scale in zip(batch, scales)
        ]
        output = self.training_model(
            self.compute_inputs(images),
            self.compute_targets(
                annotation_groups=annotation_groups,
                width=images.shape[2],
                height=images.shape[1],
            ),
        )
        if data_dir is not None:
            assert transforms is not None, "Transforms are required for data caching."
            assert indices is not None, "Image indices are required for data caching."
            os.makedirs(data_dir, exist_ok=True)
            for outputIdx, (idx, image, anns, transform, scale, metadata) in enumerate(
                zip(
                    indices or np.arange(len(images)),
                    images,
                    annotation_groups,
                    transforms,
                    scales,
                    [s.metadata for s in batch],
                )
            ):
                assert cv2.imwrite(
                    os.path.join(data_dir, str(idx) + ".png"), image[..., ::-1]
                )

                with open(
                    os.path.join(data_dir, str(idx) + ".png.metadata.json"),
                    "w",
                    encoding="utf8",
                ) as f:
                    f.write(json.dumps(metadata or {}))
                np.savez(
                    os.path.join(data_dir, str(idx) + ".png.output.npz"),
                    **{
                        k: v.detach().cpu()
                        for k, v in output["output"][outputIdx].items()
                    },
                )
                np.savez(
                    os.path.join(data_dir, str(idx) + ".png.bboxes.npz"),
                    bboxes=batch.annotation_config.bboxes_from_group(anns),
                )
                np.savez(
                    os.path.join(data_dir, str(idx) + ".png.transform.npz"),
                    transform=np.matmul(
                        np.array([[scale, 0, 0], [0, scale, 0], [0, 0, 1]]), transform
                    ),
                )
        return output["loss"]

    def train(
        self,
        training: mc.SceneCollection,
        validation: mc.SceneCollection = None,
        batch_size: int = 1,
        augmenter: mc.augmentations.AugmenterProtocol = None,
        train_backbone: bool = True,
        train_backbone_bn: bool = True,
        epochs=100,
        shuffle=True,
        callbacks: typing.List[mc.callbacks.CallbackProtocol] = None,
        optimizer_params=None,
        scheduler_params=None,
        clip_grad_norm_params=None,
        data_dir_prefix=None,
        validation_transforms: np.ndarray = None,
    ):
        """Run training job.
        Args:
            training: The collection of training images
            validation: The collection of validation images
            batch_size: The batch size to use for training
            augmenter: The augmenter for generating samples
            train_backbone: Whether to fit the backbone.
            epochs: The number of epochs to train.
            shuffle: Whether to shuffle the training data on each epoch.
            callbacks: A list of functions that accept the detector as well
                as a list of previous summaries and returns a dict of summary keys
                and values which will be added to the current summary. It can raise
                StopIteration to stop training early.
            optimizer_params: Passed to timm.optim.create_optimizer_v2 to build
                the optimizer.
            scheduler_params: Passed to timm.schduler.create_scheduler to build
                the scheduler.
        """
        training_model = (
            self.training_model if hasattr(self, "training_model") else self.model
        )
        assert training_model is not None
        optimizer = timm.optim.create_optimizer_v2(
            training_model, **(optimizer_params or DEFAULT_OPTIMIZER_PARAMS)
        )
        scheduler, _ = timm.scheduler.create_scheduler(
            types.SimpleNamespace(**(scheduler_params or DEFAULT_SCHEDULER_PARAMS)),
            optimizer=optimizer,
        )
        train_index = np.arange(len(training)).tolist()
        if validation is not None and validation_transforms is None:
            LOGGER.warning(
                "No validation transforms were provided. Assuming identity matrix."
            )
            validation_transforms = np.eye(3, 3)[np.newaxis].repeat(
                len(validation), axis=0
            )
        summaries: typing.List[typing.Dict[str, typing.Any]] = []
        for epoch in range(epochs):
            with tqdm.trange(
                len(training) // batch_size
            ) as t, tempfile.TemporaryDirectory(prefix=data_dir_prefix) as tdir:
                training_model.train()
                if not train_backbone:
                    self.freeze_backbone()
                else:
                    self.unfreeze_backbone(batchnorm=train_backbone_bn)
                t.set_description(f"Epoch {epoch + 1} / {epochs}")
                scheduler.step(
                    epoch=epoch,
                    metric=None
                    if not summaries
                    else summaries[-1].get("val_loss", summaries[-1]["loss"]),
                )
                cum_loss = 0
                for batchIdx, start in enumerate(range(0, len(training), batch_size)):
                    if batchIdx == 0 and shuffle:
                        random.shuffle(train_index)
                    end = min(start + batch_size, len(train_index))
                    train_indices = [train_index[idx] for idx in range(start, end)]
                    batch = training.assign(
                        scenes=[training[idx] for idx in train_indices]
                    )
                    if augmenter is not None:
                        batch, transforms = batch.augment(augmenter=augmenter)
                    else:
                        transforms = np.eye(3, 3)[np.newaxis].repeat(len(batch), axis=0)
                    optimizer.zero_grad()
                    loss = self.loss_for_batch(
                        batch,
                        data_dir=os.path.join(tdir, "train", str(batchIdx)),
                        transforms=transforms,
                        indices=train_indices,
                    )
                    loss.backward()
                    if clip_grad_norm_params is not None:
                        torch.nn.utils.clip_grad_norm_(
                            training_model.parameters(), **clip_grad_norm_params
                        )
                    cum_loss += loss.detach().cpu().numpy()
                    avg_loss = cum_loss / end
                    optimizer.step()
                    t.set_postfix(loss=avg_loss)
                    t.update()
                summary: typing.Dict[str, typing.Any] = {"loss": avg_loss}
                if validation is not None:
                    summary["val_loss"] = np.sum(
                        [
                            self.loss_for_batch(
                                validation.assign(
                                    scenes=[
                                        validation[idx]
                                        for idx in range(
                                            vstart,
                                            min(vstart + batch_size, len(validation)),
                                        )
                                    ]
                                ),
                                data_dir=os.path.join(tdir, "val", str(batchIdx)),
                                transforms=validation_transforms,
                                indices=np.arange(
                                    start=vstart, stop=vstart + batch_size
                                ).tolist(),
                            )
                            .detach()
                            .cpu()
                            .numpy()
                            for batchIdx, vstart in enumerate(
                                range(0, len(validation), batch_size)
                            )
                        ]
                    ) / len(validation)
                summary["lr"] = next(g["lr"] for g in optimizer.param_groups)
                if callbacks is not None:
                    try:
                        for callback in callbacks:
                            for k, v in callback(
                                detector=self,
                                summaries=summaries + [summary],
                                data_dir=tdir,
                            ).items():
                                summary[k] = v
                    except StopIteration:
                        return summaries
                t.set_postfix(**summary)
                summaries.append(summary)
        return summaries

    def detect(self, image: np.ndarray, **kwargs) -> typing.List[mc.Annotation]:
        """Run detection for a given image. All other args passed to invert_targets()

        Args:
            image: The image to run detection on

        Returns:
            A list of annotations
        """
        self.model.eval()
        images, scales = self.resize_to_model_size([image])
        with torch.no_grad():
            annotations = self.invert_targets(
                self.model(self.compute_inputs(images)), **kwargs
            )[0]
        return [a.resize(1 / scales[0]) for a in annotations]

    def detect_batch(
        self,
        images: typing.List[np.ndarray],
        batch_size: int = 32,
        **kwargs,
    ) -> typing.List[typing.List[mc.Annotation]]:
        """
        Perform object detection on a batch of images.

        Args:
            images: A list of images
            threshold: The detection threshold for the images
            batch_size: The batch size to use with the underlying model

        Returns:
            A list of lists of annotations.
        """
        self.model.eval()
        annotation_groups = []
        for start in range(0, len(images), batch_size):
            current_images, current_scales = self.resize_to_model_size(
                images[start : start + batch_size]
            )
            annotation_groups.extend(
                [
                    [a.resize(1 / scale) for a in annotations]
                    for scale, annotations in zip(
                        current_scales,
                        self.invert_targets(
                            self.model(
                                self.compute_inputs(current_images),
                            ),
                            **kwargs,
                        ),
                    )
                ]
            )
        return annotation_groups

    def mAP(
        self,
        collection: mc.SceneCollection,
        iou_threshold=0.5,
        min_threshold=0.01,
        batch_size=32,
    ):
        """Compute the mAP metric for a given collection
        of ground truth scenes.

        Args:
            collection: The collection to evaluate
            min_threshold: The minimum threshold for initial selection
                of boxes.
            iou_threshold: The IoU threshold required for
                a match

        Returns:
            mAP score
        """
        pred = collection.assign(
            scenes=[
                scene.assign(annotations=annotations)
                for scene, annotations in zip(
                    collection,
                    self.detect_batch(
                        images=collection.images,
                        threshold=min_threshold,
                        batch_size=batch_size,
                    ),
                )
            ]
        )
        return mm.mAP(
            true_collection=collection,
            pred_collection=pred,
            iou_threshold=iou_threshold,
        )

    def to_torchserve(
        self,
        model_name: str,
        directory=".",
        archive_format: tx.Literal["default", "no-archive"] = "default",
        score_threshold: float = 0.5,
        enable_flexible_size=False,
        model_version="1.0",
    ):
        """Build a TorchServe-compatible MAR file for this model."""
        assert (
            marmpu is not None
        ), "You must `pip install torch-model-archiver` to use this function."
        os.makedirs(directory, exist_ok=True)
        with tempfile.TemporaryDirectory() as tdir:
            serialized_file = os.path.join(tdir, "weights.pth")
            index_to_name_file = os.path.join(tdir, "index_to_name.json")
            model_file = os.path.join(tdir, "model.py")
            handler_file = os.path.join(tdir, "object_detector.py")
            torch.save(self.model.state_dict(prefix="model."), serialized_file)
            with open(index_to_name_file, "w", encoding="utf8") as f:
                f.write(json.dumps(self.serve_module_index))
            with open(model_file, "w", encoding="utf8") as f:
                f.write(
                    self.serve_module_string(enable_flexible_size=enable_flexible_size)
                )
            with open(handler_file, "w", encoding="utf8") as f:
                f.write(
                    pkg_resources.resource_string(
                        "mira", "detectors/assets/serve/object_detector.py"
                    )
                    .decode("utf-8")
                    .replace("SCORE_THRESHOLD", str(score_threshold))  # type: ignore
                )
            args = types.SimpleNamespace(
                model_name=model_name,
                serialized_file=serialized_file,
                handler=handler_file,
                model_file=model_file,
                version=model_version,
                requirements_file=None,
                runtime="python",
                extra_files=index_to_name_file,
                export_path=directory,
                force=True,
                archive_format=archive_format,
            )
            marmp.package_model(
                args=args, manifest=marmpu.ModelExportUtils.generate_manifest_json(args)
            )

    @property
    def anchor_sizes(self):
        """Get an array of anchor sizes (i.e., width and height)."""
        return np.diff(
            self.anchor_boxes.reshape((-1, 2, 2)),
            axis=1,
        )[:, 0, :]



""" PyTorch EfficientDet support benches

Hacked together by Ross Wightman
"""
from typing import Optional, Dict, List
import torch
import torch.nn as nn
from .anchors import Anchors, AnchorLabeler, generate_detections
from .loss import DetectionLoss


def _post_process(
    cls_outputs: List[torch.Tensor],
    box_outputs: List[torch.Tensor],
    num_levels: int,
    num_classes: int,
    max_detection_points: int = 5000,
):
    """Selects top-k predictions.

    Post-proc code adapted from Tensorflow version at: https://github.com/google/automl/tree/master/efficientdet
    and optimized for PyTorch.

    Args:
        cls_outputs: an OrderDict with keys representing levels and values
            representing logits in [batch_size, height, width, num_anchors].

        box_outputs: an OrderDict with keys representing levels and values
            representing box regression targets in [batch_size, height, width, num_anchors * 4].

        num_levels (int): number of feature levels

        num_classes (int): number of output classes
    """
    batch_size = cls_outputs[0].shape[0]
    cls_outputs_all = torch.cat(
        [
            cls_outputs[level]
            .permute(0, 2, 3, 1)
            .reshape([batch_size, -1, num_classes])
            for level in range(num_levels)
        ],
        1,
    )

    box_outputs_all = torch.cat(
        [
            box_outputs[level].permute(0, 2, 3, 1).reshape([batch_size, -1, 4])
            for level in range(num_levels)
        ],
        1,
    )

    _, cls_topk_indices_all = torch.topk(
        cls_outputs_all.reshape(batch_size, -1), dim=1, k=max_detection_points
    )
    indices_all = cls_topk_indices_all // num_classes
    classes_all = cls_topk_indices_all % num_classes

    box_outputs_all_after_topk = torch.gather(
        box_outputs_all, 1, indices_all.unsqueeze(2).expand(-1, -1, 4)
    )

    cls_outputs_all_after_topk = torch.gather(
        cls_outputs_all, 1, indices_all.unsqueeze(2).expand(-1, -1, num_classes)
    )
    cls_outputs_all_after_topk = torch.gather(
        cls_outputs_all_after_topk, 2, classes_all.unsqueeze(2)
    )

    return (
        cls_outputs_all_after_topk,
        box_outputs_all_after_topk,
        indices_all,
        classes_all,
    )


@torch.jit.script
def _batch_detection(
    batch_size: int,
    class_out,
    box_out,
    anchor_boxes,
    indices,
    classes,
    img_scale: Optional[torch.Tensor] = None,
    img_size: Optional[torch.Tensor] = None,
    max_det_per_image: int = 100,
    soft_nms: bool = False,
):
    batch_detections = []
    # FIXME we may be able to do this as a batch with some tensor reshaping/indexing, PR welcome
    for i in range(batch_size):
        img_scale_i = None if img_scale is None else img_scale[i]
        img_size_i = None if img_size is None else img_size[i]
        detections = generate_detections(
            class_out[i],
            box_out[i],
            anchor_boxes,
            indices[i],
            classes[i],
            img_scale_i,
            img_size_i,
            max_det_per_image=max_det_per_image,
            soft_nms=soft_nms,
        )
        batch_detections.append(detections)
    return torch.stack(batch_detections, dim=0)


class DetBenchPredict(nn.Module):
    def __init__(self, model):
        super().__init__()
        self.model = model
        self.config = (
            model.config
        )  # FIXME remove this when we can use @property (torchscript limitation)
        self.num_levels = model.config.num_levels
        self.num_classes = model.config.num_classes
        self.anchors = Anchors.from_config(model.config)
        self.max_detection_points = model.config.max_detection_points
        self.max_det_per_image = model.config.max_det_per_image
        self.soft_nms = model.config.soft_nms

    def forward(self, x, img_info: Optional[Dict[str, torch.Tensor]] = None):
        class_out, box_out = self.model(x)
        class_out, box_out, indices, classes = _post_process(
            class_out,
            box_out,
            num_levels=self.num_levels,
            num_classes=self.num_classes,
            max_detection_points=self.max_detection_points,
        )
        if img_info is None:
            img_scale, img_size = None, None
        else:
            img_scale, img_size = img_info["img_scale"], img_info["img_size"]
        return _batch_detection(
            x.shape[0],
            class_out,
            box_out,
            self.anchors.boxes,
            indices,
            classes,
            img_scale,
            img_size,
            max_det_per_image=self.max_det_per_image,
            soft_nms=self.soft_nms,
        )


class DetBenchTrain(nn.Module):
    def __init__(self, model, create_labeler=True):
        super().__init__()
        self.model = model
        self.config = (
            model.config
        )  # FIXME remove this when we can use @property (torchscript limitation)
        self.num_levels = model.config.num_levels
        self.num_classes = model.config.num_classes
        self.anchors = Anchors.from_config(model.config)
        self.max_detection_points = model.config.max_detection_points
        self.max_det_per_image = model.config.max_det_per_image
        self.soft_nms = model.config.soft_nms
        self.anchor_labeler = None
        if create_labeler:
            self.anchor_labeler = AnchorLabeler(
                self.anchors, self.num_classes, match_threshold=0.5
            )
        self.loss_fn = DetectionLoss(model.config)

    def forward(self, x, target: Dict[str, torch.Tensor]):
        y = self.model(x)
        class_out, box_out = [
            [
                torch.stack([yi[f"{k}_{level}"] for yi in y["output"]])
                for level in range(len(y["output"][0].keys()) // 2)
            ]
            for k in ["class_out", "box_out"]
        ]
        if self.anchor_labeler is None:
            # target should contain pre-computed anchor labels if labeler not present in bench
            assert "label_num_positives" in target
            cls_targets = [target[f"label_cls_{l}"] for l in range(self.num_levels)]
            box_targets = [target[f"label_bbox_{l}"] for l in range(self.num_levels)]
            num_positives = target["label_num_positives"]
        else:
            (
                cls_targets,
                box_targets,
                num_positives,
            ) = self.anchor_labeler.batch_label_anchors(target["bbox"], target["cls"])

        loss, class_loss, box_loss = self.loss_fn(
            class_out, box_out, cls_targets, box_targets, num_positives
        )
        return {
            "loss": loss,
            "output": y["output"],
        }


def unwrap_bench(model):
    # Unwrap a model in support bench so that various other fns can access the weights and attribs of the
    # underlying model directly
    if hasattr(model, "module"):  # unwrap DDP or EMA
        return unwrap_bench(model.module)
    elif hasattr(model, "model"):  # unwrap Bench -> model
        return unwrap_bench(model.model)
    else:
        return model



import os
import glob
import json
import typing
import shutil

try:
    import torch
except ImportError:
    torch = None  # type: ignore
import cv2
import numpy as np
import pandas as pd
import typing_extensions as tx

if typing.TYPE_CHECKING:
    from ..detectors.detector import Detector
from .. import metrics as mm
from . import scene as mcs
from . import annotation as mca
from . import utils as mcu

# pylint: disable=too-few-public-methods
class CallbackProtocol(tx.Protocol):
    """A protocol defining how we expect callbacks to behave."""

    def __call__(
        self,
        detector: "Detector",
        summaries: typing.List[typing.Dict[str, typing.Any]],
        data_dir: str,
    ) -> typing.Dict[str, typing.Any]:
        pass


def best_weights(
    filepath, metric="loss", method="min", key="saved"
) -> CallbackProtocol:
    """A callback that saves the best model weights according to a metric.

    Args:
        metric: The metric to track. Use dot notation for nested attributes
            (e.g., val_mAP.{class_name}).
        method: How to handle the metric ("min" minimizes the metric while "max"
            maximizes it).
        key: What name to use for the saved flag.
    """
    if torch is None:
        raise ValueError("You must install pytorch to use this callback.")

    # pylint: disable=unused-argument
    def callback(detector, summaries, data_dir):
        saved = False
        summaries_df = pd.json_normalize(summaries)
        best_idx = (
            summaries_df[metric].idxmax()
            if method == "max"
            else summaries_df[metric].idxmin()
        )
        if best_idx == len(summaries_df) - 1:
            torch.save(detector.model.state_dict(), filepath)
            saved = True
        return {key: saved}

    return callback


def csv_logger(filepath) -> CallbackProtocol:
    """A callback that saves a CSV of the summaries to a specific
    filepath.

    Args:
        filepath: The filepath where the logs will be saved.
    """
    # pylint: disable=unused-argument
    def callback(detector, summaries, data_dir):
        pd.json_normalize(summaries).to_csv(filepath, index=False)
        return {}

    return callback


def load_json(filepath: str):
    """Load JSON from file"""
    with open(filepath, "r", encoding="utf8") as f:
        return json.loads(f.read())


def data_dir_to_collections(data_dir: str, threshold: float, detector: "Detector"):
    """Convert a temporary training artifact directory into a set
    of train and validation (if present) true/predicted collections."""
    return {
        split: {
            "collections": {
                "true_collection": mcs.SceneCollection(
                    [
                        mcs.Scene(
                            image=filepath,
                            annotation_config=detector.annotation_config,
                            annotations=[
                                mca.Annotation(
                                    detector.annotation_config[cIdx], x1, y1, x2, y2
                                )
                                for x1, y1, x2, y2, cIdx in np.load(
                                    filepath + ".bboxes.npz"
                                )["bboxes"]
                            ],
                        )
                        for filepath in images
                    ],
                    annotation_config=detector.annotation_config,
                ),
                "pred_collection": mcs.SceneCollection(
                    [
                        mcs.Scene(
                            image=filepath,
                            annotation_config=detector.annotation_config,
                            annotations=annotations,
                        )
                        for filepath, annotations in zip(
                            images,
                            detector.invert_targets(
                                {
                                    "output": [
                                        {
                                            k: torch.Tensor(v)
                                            for k, v in np.load(
                                                filepath + ".output.npz"
                                            ).items()
                                        }
                                        for filepath in images
                                    ]
                                },
                                threshold=threshold,
                            ),
                        )
                    ],
                    annotation_config=detector.annotation_config,
                ),
            },
            "transforms": transforms,
            "metadata": pd.json_normalize(metadatas),
            "indices": [
                int(os.path.basename(filepath).split(".")[0]) for filepath in images
            ],
        }
        for split, images, metadatas, transforms in [
            (
                split,
                images,
                [load_json(f + ".metadata.json") for f in images],
                [np.load(f + ".transform.npz")["transform"] for f in images],
            )
            for split, images in [
                (split, glob.glob(os.path.join(data_dir, split, "*", "*.png")))
                for split in ["train", "val"]
            ]
        ]
        if len(images) > 0
    }


def mAP(iou_threshold=0.5, threshold=0.01) -> CallbackProtocol:
    """Build a callback that computes mAP. All kwargs
    passed to detector.mAP()"""

    # pylint: disable=unused-argument
    def callback(detector, summaries, data_dir):
        return {
            f"{split}_mAP": round(
                np.nanmean(
                    list(
                        mm.mAP(
                            **split_data["collections"],
                            iou_threshold=iou_threshold,
                        ).values()
                    )
                ),
                2,
            )
            for split, split_data in data_dir_to_collections(
                data_dir, threshold=threshold, detector=detector
            ).items()
        }

    return callback


def error_examples(
    examples_dir: str,
    threshold=0.5,
    iou_threshold=0.5,
    pad=10,
    max_crops_per_image=10,
    overwrite=False,
) -> CallbackProtocol:
    """Build a callback that saves errors to an output directory."""
    if os.path.isdir(examples_dir):
        if overwrite:
            shutil.rmtree(examples_dir, ignore_errors=True)
        else:
            raise ValueError(
                "Directory already exists. Delete it or set overwrite=True."
            )

    def callback(detector, summaries, data_dir):
        counts: typing.Dict[str, int] = {}
        entries: typing.List[dict] = []
        for split, split_data in data_dir_to_collections(
            data_dir, threshold=threshold, detector=detector
        ).items():
            for imageIdx, image, examples, metadata, transform in zip(
                split_data["indices"],
                split_data["collections"]["true_collection"].images,
                mm.crop_error_examples(
                    **split_data["collections"],
                    threshold=threshold,
                    iou_threshold=iou_threshold,
                ),
                split_data["metadata"].to_dict(orient="records")
                or [{}] * len(split_data["metadata"]),
                split_data["transforms"],
            ):
                for error_type, annotations in examples.items():
                    directory = os.path.join(
                        examples_dir, str(len(summaries)), split, error_type
                    )
                    bboxes = mcu.transform_bboxes(
                        np.array([ann.x1y1x2y2() for ann in annotations]),
                        np.linalg.pinv(transform),
                        clip=False,
                    )
                    entries.extend(
                        [
                            {
                                **metadata,
                                "split": split,
                                "imageIdx": imageIdx,
                                "annIdx": annIdx,
                                "x1": x1,
                                "y1": y1,
                                "x2": x2,
                                "y2": y2,
                                "score": ann.score,
                                "error_type": error_type,
                                "category": ann.category.name,
                            }
                            for annIdx, (ann, (x1, y1, x2, y2)) in enumerate(
                                zip(
                                    annotations,
                                    bboxes,
                                )
                            )
                        ]
                    )
                    os.makedirs(directory, exist_ok=True)
                    counts[error_type] = counts.get(error_type, 0) + len(annotations)
                    for annIdx, annotation in enumerate(
                        sorted(
                            annotations,
                            key=lambda ann: ann.score
                            if ann.score is not None
                            else ann.area(),
                            reverse=True,
                        )
                    ):
                        if annIdx >= max_crops_per_image:
                            break
                        cv2.imwrite(
                            os.path.join(directory, f"{imageIdx}_{annIdx}.png"),
                            annotation.extract(image, pad=pad)[..., ::-1],
                        )
        pd.DataFrame(entries).to_csv(
            os.path.join(examples_dir, f"{len(summaries)}.csv"), index=False
        )
        return counts

    return callback
